#ifndef MSP_GL_TEXTURE3D_H_
#define MSP_GL_TEXTURE3D_H_

#include <string>
#include "texture.h"

namespace Msp {
namespace GL {

/**
Three-dimensional texture.  Consists of an array of texels in the shape of a
right cuboid.  Texture coordinates have a principal range of [0, 1].
*/
class Texture3D: public Texture
{
public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<Texture3D, Texture::Loader>
	{
	public:
		Loader(Texture3D &);
		Loader(Texture3D &, Collection &);
	private:
		void init();

		void raw_data(const std::string &);
		void storage(PixelFormat, unsigned, unsigned, unsigned);
		void storage_levels(PixelFormat, unsigned, unsigned, unsigned, unsigned);
	};

private:
	unsigned width;
	unsigned height;
	unsigned depth;
	unsigned levels;
	unsigned allocated;

protected:
	Texture3D(GLenum);
public:
	Texture3D();

	/** Defines storage structure for the texture.  If lv is zero, the number
	of mipmap levels is automatically determined from storage dimensions.

	Must be called before an image can be uploaded.  Once storage is defined,
	it can't be changed. */
	void storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned dp, unsigned lv = 0);

	DEPRECATED void storage(PixelComponents c, unsigned w, unsigned h, unsigned d, unsigned l = 0)
	{ storage(make_pixelformat(c, UNSIGNED_BYTE), w, h, d, l); }

	/** Allocates storage for the texture.  The contents are initially
	undefined.  If storage has already been allocated, does nothing. */
	void allocate(unsigned level);

	/** Updates the contents of the entire texture.  Storage must be defined
	beforehand.  The image data must have dimensions and format matching the
	defined storage. */
	void image(unsigned level, const void *data);

	DEPRECATED void image(unsigned level, PixelComponents comp, DataType type, const void *data);

	/** Updates a cuboid-shaped region of the texture.  Storage must be defined
	beforehand.  The image data must be in a format mathing the defined storage
	and the update region must be fully inside the texture. */
	void sub_image(unsigned level, int x, int y, int z, unsigned wd, unsigned ht, unsigned dp, const void *data);

	DEPRECATED void sub_image(unsigned level,
		int x, int y, int z, unsigned wd, unsigned ht, unsigned dp,
		PixelComponents comp, DataType type, const void *data);

	/** Updates the contents of the entire texture from an image.  If storage
	has not been defined, it will be set to match the image.  In this case the
	image will be treated as a stack of square layers and its height must be
	divisible by its width.  Otherwise the image must match the defined
	storage. */
	virtual void image(const Graphics::Image &, unsigned = 0);

	using Texture::image;

	unsigned get_width() const { return width; }
	unsigned get_height() const { return height; }
	unsigned get_depth() const { return depth; }
protected:
	unsigned get_n_levels() const;
	void get_level_size(unsigned, unsigned &, unsigned &, unsigned &) const;

public:
	virtual AsyncLoader *load(IO::Seekable &, const Resources * = 0) { return 0; }
	virtual UInt64 get_data_size() const;
	virtual void unload() { }
};

} // namespace GL
} // namespace Msp

#endif
